#----------------------------------------------------------------
# Copyright (c) 2004 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------


use strict;

use Audio::Data;

package Analyze;

=head2 find_attack_simple

Look for the first sample that has a value of more than 0.025.

=cut

sub find_attack_simple {
  my $sound = shift; # Audio::Data object

  my $n = $sound->samples; # number of samples
  for (my $i=0; $i<=$n-1; $i++) {
    my $y = abs($sound->[$i]);
    if ($y>0.05) {return $i}
  }
  return undef;
}

=head2 find_attack

Look for the loudest sample that occurs after the first 50 ms (default value); the attack is only
searched for in the time before this. Next, look for the fastest increase in loudness.
Now smooth the loudness curve up to that time, using a time window of 50 ms (default value).
Then go back in time from there to until we find a channel that is softer than either
of its neighbors. Seems to work well for piano and guitar, but not for sax, which has a sustain.

=cut

sub find_attack {
  my $sound = shift; # Audio::Data object
  my $start_at = 0.050; # default to 50 ms
  if (@_) {$start_at = shift}
  my $smooth = 0.050; # default to 50 ms
  if (@_) {$smooth = shift}
  my $rate = $sound->rate;
  $start_at = int($start_at*$rate); # convert from seconds to units of the sampling period
  my $n = $sound->samples; # number of samples

  # Find the loudest sample:
  my $loudest = 0;
  my $loudest_i = $n-1;
  for (my $i=$start_at; $i<=$n-1; $i++) {
    my $y = abs($sound->[$i]);
    if ($y>$loudest) {
      $loudest = $y;
      $loudest_i = $i;
    }
  }

  # Find the fastest increase in loudness:
  my $fastest = 0;
  my $fastest_i = 0;
  for (my $i=$start_at; $i<=$loudest_i-1; $i++) {
    my $slope = abs($sound->[$i+1])-abs($sound->[$i]);
    if ($slope>$fastest) {
      $fastest = $slope;
      $fastest_i = $i;
    }
  }

  # Make a copy and smooth it:
  my @smoothed = ();
  my $window = int($smooth*$rate);
  my $k = 0; # what element of @smoothed we're about to do
  my $earliest = $fastest_i-$smooth*10;
  if ($earliest<$start_at) {$earliest=$start_at}
  for (my $i=$earliest; $i<=$fastest_i; $i++) {
    if ($k>0 && $k%100!=0) { # Do it from scratch every hundredth time to keep from accumulating rounding errors.
      $smoothed[$k] = $smoothed[$k-1]+$sound->[$i]-$sound->[$i-$window-1];
    }
    else {
      my $sum = 0;
      for (my $j=$i-$window; $j<=$i && $j<=$n-1; $j++) {
        if ($j>=0) {
          $sum = $sum+$sound->[$j];
        }
      }
      push @smoothed,$sum;
  	}
		++$k;
  }

  # Find local minimum:
  my $step = int($window/100.);
  if ($step<0) {$step=1}
  for (my $i=$fastest_i-1-$earliest-$step; $i>=0; $i-=$step) {
    if (abs($smoothed[$i])>abs($smoothed[$i+$step])) {return $i+1+$earliest}
  }
  return $earliest;

}

=head2 estimate_rms

estimate_rms($audio,$from,$to)

Optional 4th argument is number of random samples to take, default=100.
Returns a list consisting of the mean and the standard deviation.

=cut

sub estimate_rms {
  my $audio = shift;
  my $from = shift;
  my $to = shift;
  my $n = 100;
  if (@_) {$n=shift}
  my $sum_sq = 0;
  my $sum = 0;
  for (my $i=1; $i<=$n; $i++) {
    my $j = int(rand($to-$from+1))+$from;
    my $y = $audio->[$j];
    $sum_sq = $sum_sq + $y*$y;
    $sum = $sum + $y;
  }
  my $avg = $sum/$n;
  my $sd = sqrt($sum_sq/$n-$avg*$avg);
  return ($avg,$sd);
}

=head2 period_cepstrum

period_cepstrum($sound,$from,$to,$min,$max)

Returns the estimated period. Pads with silence if necessary to get a power of two length.
Only looks for periods between $min and $max. Seems to work better if I don't take log
of frequency spectrum, so this isn't a classical cepstrum.

Tested on a 523.25 Hz synthesized tone, and it worked to within a fraction of a cent.
Playing the tone through the speaker and into my electronic tuner, it was within about 5 cents,
so the tuner and the computer seem to have calibrations consistent to within 5 cents.
With real tones (guitar and tuning fork), it's much worse: systematic error of about
+25 cents, and random error of about +-20 cents.

The period_biggest_fft_peak method seems far superior.

=cut

sub period_cepstrum {
  my $sound = shift;
  my $from = shift;
  my $to = shift;
  my $min = shift;
  my $max = shift;
  my $rate = $sound->rate;
  my $part = $sound->timerange($from/$rate,$to/$rate);
  my $size = $to-$from+1;
  my $freq = $part->fft($size); # complex
  my $db = $part->clone; # real
  for (my $i=0; $i<=$size-1; $i++) {
    my $u = $freq->amplitude($i);
    $u = $u*$u;
    #$db->[$i] = log($u)+.01);
    $db->[$i] = $u;
  }
  my $cepstrum = $db->fft($size); # complex
  my $biggest = 0;
  my $biggest_i = 0;
  $min = int($min);
  $max = int($max);
  if ($min<1) {$min=1}
  if ($max>$size-1) {$max=$size-1}
  for (my $i=$min; $i<=$max; $i++) {
    my $u = $cepstrum->amplitude($i);
    if ($u>$biggest) {
      $biggest_i = $i;
      $biggest = $u;
    }
  }
  my $x2 = $biggest_i;
  my $x1 = $x2-1;
  my $x3 = $x2+1;
  my $y1 = $cepstrum->amplitude($x1);
  my $y2 = $cepstrum->amplitude($x2);
  my $y3 = $cepstrum->amplitude($x3);
  my $peak = fit_peak($x1,$x2,$x3,$y1,$y2,$y3);
  return $peak;
}

=head2 period_biggest_fft_peak

period_biggest_fft_peak($sound,$from,$to,$min,$max)

Returns the estimated period. Pads with silence if necessary to get a power of two length.
Only looks for periods between $min and $max.

Tested with tuning fork, synthesized tone, guitar notes, seems accurate to about +- 4 cents,
with no noticeable systematic error.

=cut

sub period_biggest_fft_peak {
  my $sound = shift;
  my $from = shift;
  my $to = shift;
  my $min_period = shift;
  my $max_period = shift;
  my $rate = $sound->rate;
  my $part = $sound->timerange($from/$rate,$to/$rate);
  my $size = $to-$from+1;
  my $freq = $part->fft($size); # complex
  my $biggest = 0;
  my $biggest_i = 0;
  my $min = int($rate/$max_period);
  my $max = int($rate/$min_period);
  if ($min<1) {$min=1}
  if ($max>$size-1) {$max=$size-1}
  for (my $i=$min; $i<=$max; $i++) {
    my $u = $freq->amplitude($i);
    if ($u>$biggest) {
      $biggest_i = $i;
      $biggest = $u;
    }
  }
  my $x2 = $biggest_i;
  my $x1 = $x2-1;
  my $x3 = $x2+1;
  my $y1 = $freq->amplitude($x1);
  my $y2 = $freq->amplitude($x2);
  my $y3 = $freq->amplitude($x3);
  my $peak = fit_peak($x1,$x2,$x3,$y1,$y2,$y3);
  my $freq = $rate*$peak/$size; # in Hz
  return ($rate/$freq); # i.e., size/peak
}

sub fit_peak {
  my ($x1,$x2,$x3,$y1,$y2,$y3) = @_;
  # Find slope on left and right, interpolate for zero slope.
  my $slope1 = $y2-$y1;
  my $slope2 = $y3-$y2;
  my $peak = $x2;
  if ($slope1*$slope2<0.) { # Should always pass this test, but if we don't we have a rough value for peak.
    $peak = $x2+.5*($slope2+$slope1)/($slope1-$slope2);
    if ($peak<$x1) {$peak=$x1}
    if ($peak>$x3) {$peak=$x3}
  }
  return $peak;
}

1;
