#----------------------------------------------------------------
# Copyright (c) 2004 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------


use strict;

package FiniteStateMachine;
=head1 FiniteStateMachine class

Represents a finite state machine. Transitions can be automatic, or based on
choices from the user. Reads a description of the nodes and transitions from
an input file, and rereads it every time there's a transition; this allows it
to be modified completely on the fly. There must be one node called 'start'
and one called 'end'. Errors include: (1) more
than one automatic transition out of a node; (2) automatic transitions that
form a closed loop.
Warnings include: (1) no way from 'start' to 'end' (could just indicate
a forever loop; (2) a node other than 'start' that can't be reached from
'start' (could just be a node that we've temporarily deactivated).

A node with no transitions out automatically makes a transition back
to whichever node it came from. This makes the language more expressive and
concise than a pure finite state machine.

Sample input file:

 begin office_drone;
   start> coffee*; # the * indicates an automatic transition
   coffee: meeting websurf end;
   meeting: coffee;
   websurf: coffee;
 end office_drone;

 begin male_office_drone(office_drone);
   coffee> fantasy_football;  # allow a transition from coffee to fantasy_football
   coffee> websurf-;    # don't allow transitions from coffee to websurfing
         meeting-;
         fantasy_football: coffee websurf;
 end male_office_drone;

In this example, male_office_drone is a sort of subclass of office_drone.
The stuff inside the parens is parsed first, and then we add more nodes
and/or transitions. The > is just like a :, but guards against typos
by requiring that the node already exist. The start and end nodes are
created automatically, so they should always be given with >'s.
The - says to delete that node or transition. (Deleting
a node implies deleting transitions into and out of it.)

The distinction between automatic
and nonautomatic transitions has two purposes:
(1) It helps us to avoid infinite loops. (2) As a convenience, we
automatically make the transition, whereas initiating transitions is normally
the responsibility of the external code.

Not yet implemented: deleting transitions and nodes.

=cut

=head2 new

FiniteStateMachine->new('machine',['/where/the/input/file/is','inputfile']);
FiniteStateMachine->new('machine','input_text');
The first argument gives the name of the finite state machine within the input. The input can
be given either as a directory and a file or as the literal text.
Returns a string if there's an error. 

=cut

sub new {
  my $class = shift;
  my $machine = shift;
  my $defs = shift;


  my $dir = shift;
  my $file = shift;
  my $self = {};
  bless($self,$class);
  if (ref $defs) {
    $self->{FILE} = "$defs->[0]/$defs->[1]";
  }
  else {
    $self->{DEFS} = $defs;
  }
  $self->{MACHINE} = $machine;
  return $self;
}

sub set_debug {
  my $self = shift;
  $self->{DEBUG} = shift;
}

sub debug_is_on {
  my $self = shift;
  return (exists $self->{DEBUG}) && $self->{DEBUG};
}

=head2 start

Start the finite state machine: put it in the 'start' state, and if the start state has automatic transitions out,
do them. This is not included in new(), because after doing new(), we may take some time to register all the
subs. Note that start is not allowed to have a sub.

=cut

sub start {
  my $self = shift;
  $self->{STATE} = 'start';
  my ($state,$errors,$warnings) = $self->transition(''); # in case start has automatic transitions out
  return $errors if $errors;
  return undef;
}

=head2 state

Tell the current state.

=cut

sub state {
  my $self = shift;
  return $self->{STATE};
}

=head2 defs

Return the full text of the definitions. If we were created with a pointer to
a file, then we read the file back again every time; this is a feature not a bug, 
because it allows us to change the finite state machine on the fly.

=cut

sub defs {
  my $self = shift;
  if (exists $self->{DEFS}) {return $self->{DEFS}}
  local $/; # slurp whole file
  my $file = $self->{FILE};
  open(F,"<$file") or return "error opening $file, $!";
  my $defs = <F>;
  close F;
  return $defs;
}

=head2 mama

Sets a ref that every subroutine
will get as its first argument. This is typically some mother object containing data that
all the notes need to know about.

=cut

sub mama {
  my $self = shift;
  my $mama = shift;
  $self->{MAMA} = $mama;
}

=head2 register

Takes a ref to a hash of subroutine pointers, and associates them with the
corresponding nodes of the FSM. The subroutine is called when we enter the node.
Can call this more than once, in which case the second hash is merged into the
first one.

=cut

sub register {
  my $self = shift;
  my $hash_ref = shift;
  if (!exists $self->{SUBS}) {
    $self->{SUBS} = {};
  }
  foreach my $key(%$hash_ref) {
    $self->{SUBS}->{$key} = $hash_ref->{$key};
  }
}

=head2 transition

Make a transition. If the transition is illegal, nothing is done. If the node we're
entering has an automatic transition out, we do that (and keep on going, if necessary).
Returns ($final_state,$errors,$warnings). First argument is name of state we're
going to; this can be null if all we want to do is call the sub and check for automatic transitions (only
really occurs when start has an automatic transition). Optional second arg is max number of recursions (to guard against infinite recursion
with automatic transitions). It's perfectly OK if there's no sub defined for entering this
node; that's not even a warning.

=cut

sub transition {
  my $self = shift;
  my $to = shift;
  my $max_recursions = 100;
  if (@_) {$max_recursions = shift}
  my $allowed = $self->list_transitions();
  return ($self->state,$allowed,'') if (! ref $allowed);
  my $from = $self->{STATE};
  if ($to) { # $to is null when we first create the FSM, and check for automatic transitions out of the start state
    if ((keys %$allowed) && (! exists $allowed->{$to})) { # first clause is so pop-back transitions are allowed
			return ($from,'',"illegal transition from $from to $to was not carried out");
    }
    $self->{STATE} = $to;
  }
  my $state = $self->state;
  if ($self->debug_is_on) {
    print "state is $state\n";
  }

  # trigger the sub for $state:  
  my $sub = $self->{SUBS}->{$state};
  if (defined $sub) {
    &$sub($self->{MAMA});
  }

  # now check for further, automatic transitions
  $allowed = $self->list_transitions;
  my @further = %$allowed; # flatten the hash into a list
  if (0==@further && $state ne 'end') { # pop back to wherever we came from
    return ($state,"too many recursions at node $state",'') if ($max_recursions==0);
    return $self->transition($from,$max_recursions-1);
  }
  if (2==@further && $further[1] eq 'auto') {
    return ($state,"too many recursions at node $state",'') if ($max_recursions==0);
    return $self->transition($further[0],$max_recursions-1);
  }
  return ($state,'','');
}

=head2 list_transitions

Returns a list of possible transitions out of the current node, and
in the case where there's only one, tells us if it's automatic.
Returns a string if there's an error.
The non-error return value is {node1=>1,node2=>1,...} if not automatic,
or {node=>'auto'} if automatic.

=cut

sub list_transitions {
  my $self = shift;
  my $network = {'start'=>{},'end'=>{}};
  my $defs = $self->defs();
  $defs =~ s/\#[^\n]*//g; # filter out comments
  $defs =~ s/\s+/ /g; # replace any stretch of whitespace with a single blank
  my $err = $self->parse_defs($network,$defs,$self->{MACHINE});
  return $err if $err;
  my $state = $self->{STATE};
  return "state $state not found" if (! exists $network->{$state});
  return $network->{$state};
}

sub parse_defs {
  my $self = shift;
  my $network = shift;
  my $defs = shift;
  my $machine = shift;
  if (!($defs =~ m/(.*)begin $machine ?(\([^)]+\))?;(.*)end $machine ?;/)) {
    return "definition of $machine not found";
  }
  my ($preceding,$superclass,$transitions) = ($1, $2, $3);
  if ($superclass) {
    my $err = $self->parse_defs($network,$preceding,$superclass); # no infinite recursion, because we only look backward
    return $err if $err;
  }
  while ($transitions=~m/([\w\d]+) ?([:>]) ?([^;]+)/g) {
    my ($from,$colon,$possibilities) = ($1,$2,$3);
    if (! exists $network->{$from}) {
      return "node $from was not predefined at $from$colon $possibilities" if $colon eq '>';
      $network->{$from} = {};
    }
    # strip leading and trailing blanks
    $possibilities =~ s/^ //;
    $possibilities =~ s/ $//;
    my @possibilities = split ' ',$possibilities;
    foreach my $possibility(@possibilities) {
      $possibility =~ m/^([\w\d]+)(\*)?$/;
      my ($to,$automatic) = ($1,$2);
      $network->{$from}->{$to} = ($automatic ? 'auto' : 1);
    }
  }
  return undef;
}

1;
