#----------------------------------------------------------------
# Copyright (c) 2004 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------


use strict;

package Lily;

=head2

Try to parse input written in a simple subset of Lilypond notation.
We try to ignore anything we don't understand.
Bugs/limitations: 
- assumes certain things are global (key, tempo), even if they're supposed to change during the piece
- assumes relative notation even if \relative not given
- doesn't fill in measure with tie to next bar
- music starting with a rest causes an error in Synthesize (no longer true? test)
- gets confused by nested {}

=cut

sub parse_melody {
  my $ly = shift; # input string in Lilypond notation
  my @notes = ();
  my @rhythm = ();
  my %meta = ();
  my $debug = 0;

  # Extract some metadata:  
  $ly =~ m/\\key\s+(\w+\s*(\\(major|minor))?)/;
  $meta{key} = $1; # e.g., f\major
  $meta{key} =~ m/(\w+)\s*\\/; # pick out the note itself, e.g., f
  my $tonic_note_name = $1;
  #print "-->$tonic_note_name\n";
  $tonic_note_name =~ m/(\w)(\w*)/;
  $meta{tonic} = parse_pitch($1,$2);
  if ($ly =~ m/\\tempo\s+(\d+\.?)\s*\=\s*(\d+)/) { # e.g., \tempo 8.=120
    my ($tempo_unit,$tempo_bpm) = ($1,$2);
    $meta{tempo} = $tempo_bpm / 60. / parse_duration($tempo_unit);
  }
  if ($ly=~m/\\header\s*\{([^\}]*)\}/) {
    my $header = $1;
    while ($header=~m/(\w+)\s*\=\s*\"([^"]*)\"/g) {
      $meta{"header_$1"} = $2;
    }
  }

  
  # Clean out some stuff we don't use:
  $ly =~ s/\%[^\r\n]*(\r|\n)+/\n/g; # get rid of comments
  $ly =~ s/\\header\s*\{[^\}]*\}//;

  # Clean out some more stuff (stuff we didn't clean before because it contained metadata we needed to extract first):
  $ly =~ s/\\time\s+\d+\/\d+//g;
  $ly =~ s/\\key\s+(\w+\s*(\\(major|minor))?)//g;

  # Preprocess tuplets:
  $ly =~ s@\\times\s*(\d+)/(\d+)\s*\{([^\}]*)\}@pusht$1,$2 $3 popt@g;
  my @tuplet_stack = (1,);
  my $tuplet = 1.;


  # The following is a kludge to fix a bug that made it impossible to input E# as esharp, and likewise for A#.
  # We do this after getting rid of the metadata, so all that's left should be the notes. 
  $ly =~ s/(?<![abcdefg])([ae])sharp/$1is/g;
  $ly =~ s/(?<![cdefgab])([ae])s/$1es/g;

  my $relative = 1; # really should look for \relative, but this could occur inside music...
  my $last_note;
  if ($ly=~m/\\relative\s+([cdefgab])(f|s|es|is|sharp|flat)*((,|\')*)/) {
    my ($letter,$accidental,$octave) = ($1,$2,$3);
    $last_note = parse_pitch($letter,$accidental,$octave);
    if ($debug) {print "relative, start= $letter$accidental$octave\n"}
    $ly=~s/\\relative\s+([cdefgab])(f|s|es|is|sharp|flat)*((,|\')*)//g;
  }

  $ly =~ s/\\\w+//g; # try to filter out commands we don't understand

  my $tied = 0;
  my $last_time = 1;
  while ($ly=~m/(([cdefgabr])(f|s|es|is|sharp|flat)*((,|\')*)(\d*)(\.?)\s*(\~?))|(popt)|(pusht(\d+),(\d+))/gi) {
    if ($9 || $10) {
      pop @tuplet_stack if $9;
      push @tuplet_stack,($11/$12) if $10;
      $tuplet = $tuplet_stack[-1];
      next;
    }
    my ($letter,$accidental,$octave,undef,$time,$dot,$tie) = ($2,$3,$4,$5,$6,$7,$8);
    if ($debug) {print "laotdt=$letter$accidental$octave$time$dot$tie    "}
    if (!$time) {$time = $last_time}
    $last_time = $time;
    my $midi = parse_pitch($letter,$accidental,$octave);
    if ($debug) {print "midi,last,octave=$midi,$last_note,$octave  "}
    if (defined $midi && defined $last_note) {
      while (abs($midi+12-$last_note)<abs($midi-$last_note)) {$midi+=12}
      while (abs($midi-12-$last_note)<abs($midi-$last_note)) {$midi-=12}
    }
    # In relative mode, the octave notes are actually relative to the last pitch???
    if ($relative) {
      while ($octave=~m/\'/g) {$midi+=12}
      while ($octave=~m/,/g) {$midi-=12}
	  }
    $last_note = $midi if defined $midi;
    if ($debug) {print "  midi,last=$midi,$last_note\n"}
    my $rhythm = parse_duration($time,$dot);
    if (!$tie) {
      my $n = 'r';
      if (defined $midi) {$n=$midi-60}
      my $t = 1/(1/$rhythm+$tied);
      if ($debug && $tied!=0) {print "tying, $rhythm,$tied "}
      if ($debug && !defined $midi) {print "rest"}
      push @notes,$midi;
      push @rhythm,$t/$tuplet;
      $tied = 0;
	  }
    else {
      $tied += 1/$rhythm;
    }
  }
  if ($debug) {print "\n"}
  return {'notes'=>\@notes,'rhythm'=>\@rhythm,'meta'=>\%meta};
}

sub parse_duration {
  my $rhythm = shift;
  my $dot = shift; # '', '.', '..', etc.
  my $weasel = 1.;
  while ($dot=~m/\./g) {$weasel*=2}
  $rhythm *= ($weasel)/(2*$weasel-1); # e.g., rhythm*2/3 for one dot 
  return $rhythm;
}

sub parse_pitch {
  my $letter = shift;
  my $accidental = shift;
  my $octave = shift;
  my $midi = {c=>0,d=>2,e=>4,f=>5,g=>7,a=>9,b=>11,r=>undef}->{lc($letter)};
  return $midi if (! defined $midi);
  $midi += 48 if defined $midi;
  # Turned out to be easier to do the following with the loop and the hash
  # rather than with a pattern, because s is a substring of is and es and sharp.
  my %accidentals = (s=>1,is=>1,sharp=>1,f=>-1,es=>-1,flat=>-1);
  foreach my $acc(keys %accidentals) {
    $midi += $accidentals{$acc} if $accidental eq $acc;
    $midi += 2*$accidentals{$acc} if $accidental eq $acc.$acc;
  }
  while ($octave=~m/\'/g) {$midi+=12}
  while ($octave=~m/,/g) {$midi-=12}
  return $midi;
}

1;
