#----------------------------------------------------------------
# Copyright (c) 2004 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------


use strict;

package Music;

=head1 Chord class

Represents a set of pitches (no time or rhythm information). Each pitch
is represented by an integer. Can also be used to represent other sets of
pitches, e.g., scales.

=cut

package Chord;

=head2 new

Chord->new(@notes); # array of integers, which need not be sorted by pitch on input

=cut

sub new {
  my $class = shift;
  my @notes = ();
  if (@_) {@notes = @_}
  my $self = {};
  bless($self,$class);
  $self->{PITCHES} = \@notes; # can't call notes() method, because it fails when array has no elements
  return $self;
}

sub scale {
  my $self = shift;
  my $type = shift;
  my $tonic = 0;
  if (@_) {$tonic = shift}
  my %rel_pitches = 
  (
    'major'              =>[0,2,4,5,7,9,11],
    'natural_minor'      =>[0,2,3,5,7,8,10],
    'harmonic_minor'     =>[0,2,3,5,7,8,11],
    'jazz_melodic_minor' =>[0,2,3,5,7,9,11],
	 );
  if (!exists $rel_pitches{$type}) {die "unknown scale, $type"}
  my $rel_pitches = $rel_pitches{$type};
  $self->pitches(@$rel_pitches);
  $self->transpose($tonic);
}

sub range {
  my $self = shift;
  my @pitches = $self->pitches();
  my $lowest = 9999;
  my $highest = -9999;
  foreach my $pitch(@pitches) {
    $lowest = $pitch if ($pitch<$lowest);
    $highest = $pitch if ($pitch>$highest);
  }
  return [$lowest,$highest];
}

sub transpose {
  my $self = shift;
  my $transposition = shift;
  my @pitches = $self->pitches;
  my @new_pitches = ();
  foreach my $pitch(@pitches) {
    push @new_pitches,$pitch+$transposition;
  }
  $self->pitches( @new_pitches);
}

=head2 pitches

Sets or returns an array of integers. Guaranteed to be sorted by pitch on output. Need
not be sorted by pitch on input.

=cut

sub pitches {
  my $self = shift;
  if (@_) {
    my @input = @_;
    @input = sort { $a <=> $b } @input;
    $self->{PITCHES}=\@input;
  }
  my $ref = $self->{PITCHES};
  return @$ref;
}

package Util;

# bug: only works if $tonic is 40; otherwise top note of each chord is garbage

sub random_cadence {
  my $mode = shift; # 'major' or 'minor'
  my $tonic = shift;
  my $transposition = $tonic-40;
  my ($tonic_third,$dom_ninth);
  my $major = ($mode eq 'major');
  if ($major) {
    $tonic_third = $tonic+16;
    $dom_ninth = $tonic+20; # 9th of V = 6th of tonic
  }
  else {
    $tonic_third = $tonic+15;
    $dom_ninth = $tonic+19;
  }

  my @cadences = (
    [
      Chord->new(47,51,57,66), # like GBFD
      Chord->new(40,47,$tonic_third,64), # like CGEC
    ],
    [
      Chord->new(47,57,63), # like GFB
      Chord->new(40,$tonic_third,64), # like CEC
    ],
    [
      Chord->new(47,51,57), # like GBF
      Chord->new(40,52,$tonic_third), # like CCE
    ],
    [
      Chord->new(47,51,57), # like GBF
      Chord->new(52,$tonic_third), # like CE
    ],
    [
      Chord->new(51,54,57,$dom_ninth), # like BDFA
      Chord->new(52,$tonic_third,59), # like CEG
    ],
  );

  my $cadence = random_array_element(@cadences);
  my ($chord1,$chord2) = @$cadence;
  $chord1->transpose($transposition);
  $chord2->transpose($transposition);
  if (0) {      print join(' ',$chord1->pitches) . '-->' .
									join(' ',$chord2->pitches) ."\n";}
  return ($chord1,$chord2);
}

sub random_voicing {
  my $root = shift;
  my $type = shift; # 'maj','m','maj7','7','m7','mmaj7','m7b5','o7';
  my $can_omit = shift; # string containing characters '1', '5', etc.
  my $chord;
  my ($third,$fifth,$seventh) = (4,7,11);
  if ($type eq '7') {
    $seventh = 10;
  }
  if ($type eq 'm7' || $type eq 'm') {
    $third = 3;
    $seventh = 10;
  }
  if ($type eq 'mmaj7') {
    $third = 3;
  }
  if ($type eq 'm7b5') {
    $third = 3;
    $fifth = 6;
    $seventh = 10;
  }
  if ($type eq 'o7') {
    $third = 3;
    $fifth = 6;
    $seventh = 9;
  }
  if ($type eq 'maj' || $type eq 'm' || (rand()<.5 && $can_omit=~m/7/)) { # do a bare triad
    $chord = random_array_element(
			Chord->new(0,$third,$fifth),
			Chord->new(0,$fifth,$third+12),
    );
  }
  if (!defined $chord && $type eq 'maj7' && $can_omit=~m/7/ && rand()<.5) {
    $chord = random_array_element(
      # For maj7 and mmaj7 chords, we may actually voice it as a 6th chord:
			Chord->new(0,$third,$fifth,9),
			Chord->new(0,$fifth,9,$third+12),
			Chord->new(0,9,$third+12,$fifth+12),
			Chord->new(0,9,$fifth+12,$third+24),
			Chord->new(0,$fifth,$third+12,9+12),
			Chord->new(0,$third,9,$fifth+12),
    );
    
  }
  if (!defined $chord) {
    $chord = random_array_element(
			Chord->new(0,$third,$fifth,$seventh),
			Chord->new(0,$fifth,$seventh,$third+12),
			Chord->new(0,$seventh,$third+12,$fifth+12),
			Chord->new(0,$seventh,$fifth+12,$third+24),
			Chord->new(0,$fifth,$third+12,$seventh+12),
			Chord->new(0,$third,$seventh,$fifth+12),
    );
  }
  my ($lo,$hi) = array_lo_and_hi($chord->pitches());
  # optional doublings:
  if (rand()<.3) {
    $chord->pitches($chord->pitches(),-12);
    if (rand()<.3 && $fifth==7) {
      $chord->pitches($chord->pitches(),-12+$fifth);
    }
  }
  if (rand()<.3 && !contained_in_array(11,$chord->pitches())) {
    $chord->pitches($chord->pitches(),12);
  }
  if (rand()<.2 && !contained_in_array(23,$chord->pitches()) && $hi>=12+$fifth) {
    $chord->pitches($chord->pitches(),24);
  }
  $chord->transpose($root);
  return $chord;
}

sub contained_in_array {
  my $x = shift;
  my @array = @_;
  foreach my $y(@array) {
    return 1 if $x==$y;
  }
  return 0;
}

sub random_array_index {
  my @array = @_;
  my $n = @array;
  return int(rand($n));
}

sub random_array_element {
  my @array = @_;
  my $n = @array;
  return $array[int(rand($n))];
}

sub random_array_element_weighted {
  my $hash = shift;
  my @array = keys %$hash;
  my $n = @array;
  my $norm = 0;
  foreach my $element(@array) {
    $norm += $hash->{$element};
  }
  my $x = rand($norm);
  foreach my $element(@array) {
    $x -= $hash->{$element};
    return $element if $x<0;
  }
  return $array[0]; # should never actually happen, but conceivably could because of rounding
}

sub array_lo_and_hi {
  my @array = @_;
  my ($lo,$hi);
  foreach my $x(@array) {
    $lo = $x if (defined $x && ((!defined $lo) || $x<$lo));
    $hi = $x if (defined $x && ((!defined $hi) || $x>$hi));
  }
  return ($lo,$hi);
}

sub random_integer_in_range {
  my $lo = shift;
  my $hi = shift;
  return undef if $lo>$hi;
  return int(rand($hi-$lo+1))+$lo;
}

sub add_constant_to_array {
  my $k = shift;
  my @a = @_;
  for (my $i=0; $i<=$#a; $i++) {
    $a[$i] += $k if defined $a[$i];
  }
  return @a;
}

sub interval_to_name {
  my $interval = shift;
  my @interval_names = ('u','m2','M2','m3','M3','4','t','5','m6','M6','m7','M7','8');
  while ($interval>$#interval_names) {$interval-=12}
  while ($interval<0) {$interval+=12}
  return $interval_names[$interval];
}

sub name_to_interval {
  my $name = shift;
  for (my $i=0; $i<=12; $i++) {
    if (interval_to_name($i) eq $name) {return $i};
  }
  return undef;
}

sub list_interval_names {
  my @names = ();
  for (my $i=0; $i<=12; $i++) {
    push @names,interval_to_name($i);
  }
  return join ' ',@names;
}

sub note_to_crowell { # my idiosyncratic solfeg system
  my $note = shift;
  my @solfeg = ('do','gu','ri','bu','mi','fo','ka','sa','ja','li','pa','te');
  return $solfeg[$note%12];
}

sub crowell_to_note {
  my $name = shift;
  $name = lc($name);
  $name = substr($name,0,1); # only first character is significant for this purpose
  for (my $i=0; $i<=11; $i++) {
    my $x = note_to_crowell($i);
    $x = substr($x,0,1);
    if ($x eq $name) {return $i};
  }
}

sub list_crowell_names {
  my @names = ();
  for (my $i=0; $i<=11; $i++) {
    my $name = note_to_crowell($i);
    push @names,substr($name,0,1);
  }
  return join ' ',@names;
}

sub random_octave_transposition_list {
  my $notes = shift; # array ref;
  my $range = shift; # array ref
  my $prob = .3;
  if (@_) {$prob = shift}
  my $delta = 0;
  while (rand()<$prob) {$delta += 12}
  while (rand()<$prob) {$delta -= 12}
  for (;;) {
    my $ok = 1;
    foreach my $note(@$notes) {
      $ok = $ok && (($note+$delta)<=$range->[1]);
    }
    last if $ok;
    $delta -= 12;
  }
  for (;;) {
    my $ok = 1;
    foreach my $note(@$notes) {
      $ok = $ok && (($note+$delta)>=$range->[0]);
    }
    last if $ok;
    $delta += 12;
  }
  my @result = ();
  foreach my $note(@$notes) {
    push @result,$note+$delta;
  }
  return @result;
}

sub random_octave_transposition {
  my $note = shift;
  my $range = shift; # array ref
  my $prob = .3;
  if (@_) {$prob = shift}
  while (rand()<$prob) {$note += 12}
  while (rand()<$prob) {$note -= 12}
  while ($note<$range->[0]) {$note +=12}
  while ($note>$range->[1]) {$note -=12}
  return $note;  
}

# can also supply refs to scalars in the list
sub range_of_chord_list {
  my @chords = @_;
  my ($lo,$hi);
  foreach my $chord(@chords) {
    my $range;
    if (ref($chord) eq 'Chord') {
      $range = $chord->range;
    }
    else {
      $range = [$$chord,$$chord];
    }
    $lo = $range->[0] if (!defined $lo) || ($range->[0])<$lo;
    $hi = $range->[1] if (!defined $hi) || ($range->[1])>$hi;
  }
  return ($lo,$hi);
}

# can also supply refs to scalars in the list
sub transpose_all {
  my $transposition = shift;
  my @chords = @_;
  foreach my $chord(@chords) {
    if (ref($chord) eq 'Chord') {
      $chord->transpose($transposition);
    }
    else {
      $$chord += $transposition;
    }
  }
}

sub bring_within_octave {
  my $note = shift;
  my $bring_close_to = shift;
  while ($note<=$bring_close_to-12) {$note+=12}
  while ($note>=$bring_close_to+12) {$note-=12}
  return $note;
}

# use a heuristic algorithm to try to recognize notes that begin phrases
# to do:
#    - does it work when 1st note of piece begins on barline?
#    - using -1's is goofy; don't depend on them
#    - take 5th param saying whether this is beginning of piece
#    - test systematically
sub phrase_score {
  my $timesig = shift; # number of wholenotes in a bar
  my $notes = shift; # array ref, contains undefs for rests
  my $rhythm = shift; # array ref, e.g., .25 for a quarter note
  my $bar = shift; # array ref, time since last barline (is -1 for first bar of the whole piece)
  my $how_much_i_love_the_barline = 5; # This is the most important tunable parameter. If this is 0, we still prefer the barline.
  my @rhythm = @$rhythm;
  my @bar = @$bar;
  my $n = @rhythm;
  my @score = ();
  for (my $i=0; $i<=$n-1; $i++) {
    push @score,0.;
  }
  if ($n==0) {return @score}
  # Not a note but a rest: -1000000 points
  for (my $i=0; $i<=$n-1; $i++) {
    if (! defined $notes->[$i]) {
      $score[$i] -= 1000000;
    }
  }
  # First bar of the piece: up to 10000 points
  for (my $i=0; $i<=$n-1; $i++) {
    if ($bar[$i] == -1) {
      $score[$i] += 5000; 
      if ($i==0) {
        $score[$i] += 5000;
      }
    }
  }
  # First note after a long rest: up to 100 points
  for (my $i=1; $i<=$n-1; $i++) {
    my $bonus = 0.;
    for (my $j=$i-1; $j>=0 && ! defined $notes->[$j]; $j--) {
      $bonus += 100.*$rhythm[$j];
    }
    $bonus = 100. if $bonus>100.;
    $score[$i] += $bonus;
  }
  # Lies exactly on a barline: $how_much_i_love_the_barline points
  for (my $i=0; $i<=$n-1; $i++) {
    my $bar = $bar[$i];
    if ($bar==0.) {
      $score[$i] += $how_much_i_love_the_barline;
    }
  }
  # Lies on or close to a barline: up to 20 points
  for (my $i=0; $i<=$n-1; $i++) {
    my $bar = $bar[$i];
    if ($bar>=0.) { # -1 indicates first bar of piece
      my $bonus1 = 20.*.5*$bar/$timesig;
      my $bonus2 = 20.*.5*($timesig-$bar)/$timesig;
      if ($bonus1>$bonus2) {
        $score[$i] += $bonus1;
      }
      else {
        $score[$i] += $bonus2;
      }
    }
  }  
  # Lies at a location that bears some simple relation to the barline: up to 10 points (can be negative)
  # This is meant to exactly cancel out the preference for being close to the barline in the case where
  # we just have 4/4 in quarter notes.
  for (my $i=0; $i<=$n-1; $i++) {
    my $bar = $bar[$i];
    if ($bar>=0.) { # -1 indicates first bar of piece
      $score[$i] += 10.-5.*weirdness_of_relation_to_barline($timesig,$bar);
    }
  }  
  # Has a leap before it and a step after it: 2 points
  for (my $i=1; $i<=$n-2; $i++) {
    my ($a,$b,$c) = ($notes->[$i-1],$notes->[$i],$notes->[$i+1]);
    if ((defined $a) && (defined $b) && (defined $c) && abs($a-$b)>2 && abs($b-$c)<=2) {
      $score[$i] += 2.;
    }
  }  
  return @score;
}

# ranges from 0 to 4
sub weirdness_of_relation_to_barline {
  my $timesig = shift; # wholenotes per bar
  my $bar = shift; # time since last barline
  if ($bar>.5*$timesig) {return weirdness_of_relation_to_barline($timesig,$timesig-$bar)}
  my ($a,$b,$c,$d) = float_to_prime_factorization($bar);
  if (! defined $a) {return 4.}
  my $sump = abs($a)+abs($b)+abs($c)+abs($d)-1; # =0 if 1/2 note away from the barline, =1 if 1/4 note away, ...
  if ($sump<0.) {$sump=0.}
  if ($sump>2.) {$sump=2.+($sump-2.)**.25}
  if ($sump>4.) {$sump=4.}
  return $sump;
}

sub float_to_fraction {
  my $x = shift;
  my $try = 9; # maximum power of 2 to try (uses smaller maximums for the powers of 3, 5, and 7); never use less than 9 here
  if (@_) {$try = shift}
  my ($a,$b,$c,$d) = float_to_prime_factorization($x,$try);
                      my ($u,$v) = (1,1);
											if ($a<0) {$v=$v*2**-$a}
											if ($a>0) {$u=$u*2**$a}
											if ($b<0) {$v=$v*3**-$b}
											if ($b>0) {$u=$u*3**$b}
											if ($c<0) {$v=$v*5**-$c}
											if ($c>0) {$u=$u*5**$c}
											if ($d<0) {$v=$v*7**-$d}
											if ($d>0) {$u=$u*7**$d}
                      return ($u,$v);
}

# analyze the number using the form 2**a 3**b 5**c 7**d
sub float_to_prime_factorization {
  my $x = shift;
  my $try = 9; # maximum power of 2 to try (uses smaller maximums for the powers of 3, 5, and 7); never use less than 9 here
  if (@_) {$try = shift}
  my $e = .00001;
  return undef if abs($x)<$e;
  return undef if $x<0.;

  # We have prime factorizations to search through, like 2**a*3**b*5**c*7**d,
  # and if a...d can be large numbers, the search space can be very large.
  # Try to search the most likely parts of the space first.

  for (my $maxa=1; $maxa<=$try; $maxa++) { # max |a|
    my $maxb = $maxa-7; # max |b|
    $maxb=0 if $maxb<0;
    my $maxc = $maxa-7; # max |c|
    $maxc=0 if $maxc<0;
    my $maxd = $maxa-7; # max |d|
    $maxd=0 if $maxd<0;
    for (my $absa=0; $absa<=$maxa; $absa++) {
      for (my $signa= -1; $signa<=1; $signa+=2) {
        my $a=$absa*$signa;
        for (my $absb=0; $absb<=$maxb; $absb++) {
          for (my $signb= 1; $signb>= -1; $signb-=2) { # most common value of b is 0, followed by +1 (dotted rhythm)
            my $b=$absb*$signb;
            for (my $absc=0; $absc<=$maxc; $absc++) {
              for (my $signc=-1; $signc<=1; $signc+=2) { # most common value of c is 0, followed by -1 (five over four)
                my $c=$absc*$signc;
                for (my $absd=0; $absd<=$maxd; $absd++) {
                  for (my $signd=-1; $signd<=1; $signd+=2) { # most common value of d is 0, followed by -1 (double-dotted rhythm)
                    my $d=$absd*$signd;
										return ($a,$b,$c,$d) if (abs($x-2**$a*3**$b*5**$c*7**$d)<$e);
                  }
                } # absd
              }
            } # absc
          }
        } # absb
      }
    } # absa
  }
  return undef;
}



1;
