#----------------------------------------------------------------
# Copyright (c) 2004 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------


use strict;

use Audio::Data;
use Music;
use Lily;
use Clone;
use OggToAu;

package Synthesize;
=head1 Timbre class

Represents a set of notes sampled from an acoustic instrument.

=cut

package Timbre;

=head2 new

Timbre->new('/where/they/are','ss*.au',$allow_term_output);
Returns a string if there's an error. If the .au files don't exist but .ogg files do, we silently convert them.
The first argument is a directory.
The second argument is a pattern for making filenames from 3-digit midi note numbers.
The third one says whether the subroutine is allowed to give terminal output if it has
to convert the ogg files (default=1).

=cut

sub new {
  my $class = shift;
  my $dir = shift;
  my $pat = shift;
  my $allow_term_output = 1;
  if (@_) {$allow_term_output = shift}
  my $self = {};
  bless($self,$class);
  $self->{DIR} = $dir;
  $self->{PAT} = $pat;
  my $info_file = "$dir/timbre";
  open(F,"<$info_file") or return "error opening $info_file, $!";
  my %info = ();
  while (my $line=<F>) {
    if ($line=~m/^([^=]+)\=(.*)$/) {
      my ($key,$value) = ($1,$2);
      $info{$key} = $value;
    }
  }
  close F;
  $self->{INFO} = \%info;
  if (0) {
    my $lowest = 9999;
    my $highest = -9999;
    $pat =~ m/(.*)\*(.*)/;
    my ($start_pat,$end_pat) = ($1,$2);
    my @files = glob "$dir/$pat";
    if (!@files) {return "error creating timbre, no files found matching $dir/$pat"}
    foreach my $file(@files) {
      if ($file =~ m|$dir/$start_pat(\d+)$end_pat|) {
        my $midi = $1;
        $lowest = $midi if ($midi<$lowest);
        $highest = $midi if ($midi>$highest);
    	}
    }
  }
  my $range = $info{range};
  $range =~ m/(\d+)\-(\d+)/;
  my ($lo,$hi) = ($1,$2);
  $self->{RANGE} = [$lo,$hi];
  my $best = $info{best};
  $best =~ m/(\d+)\-(\d+)/;
  $self->{BEST} = [$1,$2];
  $self->{RATE} = $info{rate};
  
  if (! -e $self->file_for_note($lo)) {
    # Do we need to convert .oggs to .aus?
    for (my $note=$lo; $note<=$hi; $note++) {
      my $au = $self->file_for_note($note);
      my $ogg = $au;
      $ogg =~ s/\.au$/\.ogg/;
      if ((! -e $au) && (! -e $ogg)) {return "error, neither $au nor $ogg exists"}
      if ((! -e $au) && -e $ogg) {
        if ($allow_term_output) {print "Converting $ogg to $au\n"}
				OggToAu::convert($ogg,$au);
      }
    }
  }

  return $self;
}

sub file_for_note {
  my $self = shift;
  my $pitch = shift; # midi
  my $dir = $self->{DIR};
  my $pat = $self->{PAT};
  my $midi = sprintf '%03d',($pitch);
  my $file = $pat;
  $file =~ s/\*/$midi/;
  return "$dir/$file";
}

sub random_transposition {
  my $self = shift;
  my $lo = shift;
  my $hi = shift;
  my $r = $self->best;
  my $transposition = Util::random_integer_in_range($r->[0]-$lo,$r->[1]-$hi);
  if (! defined $transposition) {
    $r = $self->range;
    $transposition = Util::random_integer_in_range($r->[0]-$lo,$r->[1]-$hi);
  }
  return $transposition;
}

sub random_octave_transposition {
  my $self = shift;
  my $lo = shift;
  my $hi = shift;
  my $r = $self->best;
  my $transposition = Util::random_integer_in_range($r->[0]-$lo,$r->[1]-$hi);
  if (! defined $transposition) {
    $r = $self->range;
    $transposition = Util::random_integer_in_range($r->[0]-$lo,$r->[1]-$hi);
  }
  my $mod = $transposition % 12;
  if ($mod<=6) {
    $transposition -= $mod;
  }
  else {
    $transposition += (12-$mod);
  }
  return $transposition;
}

=head2 range()

Returns a ref to a 2-element array containing the highest and lowest midi notes we can play with this timbre.

=cut

sub range {
  my $self = shift;
  return $self->{RANGE};
}

=head2 best()

Returns a ref to a 2-element array containing the highest and lowest midi notes that sound good in this timbre.

=cut

sub best {
  my $self = shift;
  return $self->{BEST};
}

sub synthesize {
  my $self = shift;
  my %args = (
               CHORDS=>[], # ref to an array of Chord objects or integers representing midi values
               RHYTHM=>[], # ref to an array like, e.g., .25 .25 .5 for two quarter notes followed by a half note
               PROPERTIES=>[], # staccato accents are 's'
               NOTE_VOLUME=>undef, # multiplies the global volume setting
               VOLUME=>1, # linear (not logarithmic); further modulated by STYLE
               TEMPO=>1, # whole notes per second
               @_,
						 );
  my $raw_chords = $args{CHORDS};
  my $raw_rhythm = $args{RHYTHM};
  my $raw_properties = $args{PROPERTIES};
  my $tempo = $args{TEMPO};
  my $styles = $args{STYLES}; 
  my $volume = $args{VOLUME};
  my $note_volume = $args{NOTE_VOLUME};
  my $rate = $self->{RATE};
  my $sound = Audio::Data->new(rate=>$rate);
  my $this_chord_sound = Audio::Data->new(rate=>$rate);
  my $null_sound = Audio::Data->new(rate=>$rate);
  $null_sound->silence(0);
  my $save_tail = $null_sound->clone; # To get a legato feel, we save the second half the the previous note's decay and overlap it with the next note.
  die "Illegal tempo, zero or undefined" if !$tempo;
  # skip initial rests, because they confuse me when trying to find rate from first sample read (no longer need to skip?)
  my $start = 0;
  my @chords = ();
  my @rhythm = ();
  my @vol = ();
  my @properties = ();
  for (my $i=0; $i<@$raw_chords; $i++) {
    my ($c,$r,$v) = ($raw_chords->[$i],$raw_rhythm->[$i],$note_volume->[$i]);
    if (!($raw_properties->[$i]=~m/s/)) {
      push @chords,$c;
      push @rhythm,$r;
      push @vol,$v;
    }
    else {
      push @chords,$c;
      push @rhythm,$r/2.;
      push @vol,$v;
      push @chords,undef;
      push @rhythm,$r/2.;
      push @vol,undef;
    }
  }
  while ($start<@chords-1 && ((!ref($chords[$start]) && !defined $chords[$start]))) {++$start}
  for (my $i=0; $i<@chords; $i++) {
    my $chord = $chords[$i];
    my $duration = $rhythm[$i]/$tempo; # seconds
    my @pitches;
    if (ref($chord)) {
      @pitches = $chord->pitches; # a Chord object
    }
    else {
      @pitches = ($chord,); # integer representing a midi note
    }
    foreach my $pitch(@pitches) {
      my $file = $self->file_for_note($pitch);
      my $this_note_sound;
      if (defined $pitch && $pitch>0) {
        $this_note_sound = get_audio_from_au_file($file);
        if (! ref $this_note_sound) {
          print "Error: $this_note_sound\n";
        }
      }
      else {
        $this_note_sound = Audio::Data->new(rate=>$rate);
        $this_note_sound->silence($duration);
      }
      $this_note_sound *= $vol[$i] if defined $vol[$i];
      $this_chord_sound = $this_chord_sound + $this_note_sound;
    }
    if (($save_tail->duration)>$duration) {
      $save_tail = $save_tail->timerange(0,$duration);
    }
    my $decay = 0.05; # seconds
    my $sustain = $duration-$decay/2;
    my $tail1 = $decay/2; # the part of the decay before the next note starts
    my $tail2 = $decay/2; # ...and the rest
    if ($duration<$decay/2) { # note too short to fit the decay
      $sustain = 0;
      $tail1 = $duration;
      $tail2 = $decay-$duration;
    }
    my $tail = $this_chord_sound->timerange($sustain,$sustain+$tail1+$tail2);
    my $nt = $tail->samples;
    my $k = 1./$nt;
    for (my $j=0; $j<=$nt; $j++) {
      $tail->[$j] *= 1.-$j*$k;
    }
    $sound = $sound . ($save_tail + $volume*(
                                               ($this_chord_sound->timerange(0.,$sustain))
                                              .($tail->timerange(0.,$tail1))
                                            )
                      );
    $save_tail = $volume*$tail->timerange($tail1,$tail1+$tail2);
    $this_chord_sound = $null_sound->clone; # for next time through loop
  }
  return $sound.$save_tail;
}

sub play_lily {
  my $self = shift;
  my $ly_file = shift;
  my $server = shift;
  my $tempo = shift;
  my $parsed = Lily::parse_melody($ly_file);
  my $notes = $parsed->{notes};
  my $rhythm = $parsed->{rhythm};
  my ($lo,$hi) = Util::array_lo_and_hi(@$notes);
  my $r = $self->range;
  my $transposition = Util::random_integer_in_range($r->[0]-$lo,$r->[1]-$hi);
  if (! defined $transposition) {die "range of melody too wide, file $ly_file, $lo to $hi"}
  @$notes = Util::add_constant_to_array($transposition,@$notes);
  my $sound = $self->synthesize(CHORDS=>$notes,RHYTHM=>$rhythm,VOLUME=>1./10.,TEMPO=>$tempo);
  $server->play($sound);

}

sub get_audio_from_au_file {
  my $filename = shift;
  my $audio = Audio::Data->new(); # normally you can't do new() without a rate, but I assume here it's going to read it from the file
  open(FH,"<$filename");
  if (! -e $filename) {return "file $filename doesn't exist"}
  binmode FH;
  $audio->Load(*FH);
  close FH;
  return $audio;
}

1;
