#----------------------------------------------------------------
# Copyright (c) 2004 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------

use strict;

package TonalActivity;

our @ISA = 'Activity';

=head1 TonalActivity class

This subclass of Activity is an abstract class, which is subclassed to make more specific types of activities.

=cut

=head2 new

TonalActivity->new(...,FSM=>'my_fsm',FSM_FILE=>['/where/it/is','my_fsm.logic'])

where ... is stuff for the Activity class.

=cut

sub new {
  my $class = shift;
  my %args = (
    FSM=>'TonalActivity',
    FSM_FILE=>['/usr/local/share/clamor/logic','TonalActivity.logic'],
		VOLUMES=>{'cadence'=>.5},
		MODE_WEIGHTS=>{'major'=>3,'minor'=>1},
    MINOR_SCALE_WEIGHTS=>{'natural_minor'=>1,'harmonic_minor'=>1,'jazz_melodic_minor'=>1},
    SAME_KEY_PROB=>.65,
    SET_KEY_MYSELF=>1,
		@_,
	);
  my $self = Activity->new(%args);
  $self->{VOLUMES} = $args{VOLUMES};
  $self->{MODE_WEIGHTS} = $args{MODE_WEIGHTS};
  $self->{MINOR_SCALE_WEIGHTS} = $args{MINOR_SCALE_WEIGHTS};
  $self->{SAME_KEY_PROB} = $args{SAME_KEY_PROB};
  $self->{SET_KEY_MYSELF} = $args{SET_KEY_MYSELF};
  bless($self,$class); # rebless into subclass
  $self->{FSM} = FiniteStateMachine->new($args{FSM},$args{FSM_FILE});
  if (! ref $self->{FSM}) {die $self->{FSM}}
  $self->{FSM}->register({
    key=>sub {
      if ($self->{SET_KEY_MYSELF}) {
        my $mode = Util::random_array_element_weighted($self->{MODE_WEIGHTS});
        my $tonic = 40;
        my ($chord1,$chord2) = Util::random_cadence($mode,$tonic);
        $self->{EXERCISE}->{CADENCE_SOUND} = undef;
        my $transposition = $self->{TIMBRES}->{accomp}->random_transposition(Util::range_of_chord_list($chord1,$chord2));
        die "chord can't be played, instrument has insufficient range" if ! defined $transposition;
        Util::transpose_all($transposition,\$tonic,$chord1,$chord2);
        $self->{EXERCISE}->{TONIC} = $tonic;
        $self->{EXERCISE}->{CADENCE} = [$chord1,$chord2];
        $self->{EXERCISE}->{MODE} = $mode;
		  }
    },
    cadence=>sub {
      $self->play_cadence();
    },
    tonic=>sub {
      my $timbre = $self->{TIMBRES}->{'melody'};
      $self->play($timbre->synthesize(CHORDS=>[$self->{EXERCISE}->{TONIC},],RHYTHM=>[1,],VOLUME=>1./2.)),
      #print "tonic=".($self->{EXERCISE}->{TONIC})."\n";
    },
    next=>sub {
      if (rand()<$self->{SAME_KEY_PROB}) {
        $self->{FSM}->transition('construct');
      }
      else {
        $self->{FSM}->transition('key');
      }
    },
  });
  return $self;
}

sub play_cadence {
  my $self = shift;

  #print "I chord=".(join ' ',($self->{EXERCISE}->{CADENCE}->[1]->pitches))."\n";
  if (! defined $self->{EXERCISE}->{CADENCE_SOUND}) {
    $self->{EXERCISE}->{CADENCE_SOUND} =  $self->{TIMBRES}->{'accomp'}->synthesize(CHORDS=>$self->{EXERCISE}->{CADENCE},
                                                                                 RHYTHM=>[1,1],VOLUME=>$self->{VOLUMES}->{cadence});
  }
  $self->play($self->{EXERCISE}->{CADENCE_SOUND});
}

# use this if SET_KEY_MYSELF is false
sub set_tonic_and_mode {
  my $self = shift;
  my $tonic = shift;
  my $mode = shift;
  my ($chord1,$chord2) = Util::random_cadence($mode,40);
  Util::transpose_all(($tonic-40),$chord1,$chord2);
  $self->{EXERCISE}->{CADENCE_SOUND} = undef;
  my $transposition = $self->{TIMBRES}->{accomp}->random_octave_transposition(Util::range_of_chord_list($chord1,$chord2));
  die "chord can't be played, instrument has insufficient range" if ! defined $transposition;
  Util::transpose_all($transposition,$chord1,$chord2);
  $self->{EXERCISE}->{TONIC} = $tonic;
  $self->{EXERCISE}->{CADENCE} = [$chord1,$chord2];
  $self->{EXERCISE}->{MODE} = $mode;
}

sub random_scale_as_array {
  my $self = shift;
  my $mode = $self->{EXERCISE}->{MODE};
  my $scale_name;
  if ($mode eq 'major') {
    $scale_name = 'major';
  }
  else {
    $scale_name = Util::random_array_element_weighted($self->{MINOR_SCALE_WEIGHTS});
  }
  my $scale = Chord->new();
  $scale->scale($scale_name,$self->{EXERCISE}->{TONIC});
  return $scale->pitches;
}


1;
